(* AUTEURS / AUTHORS :                          *)
(* Judicaël Courant,   Jean-François Monin      *)
(* VERIMAG, Université Joseph Fourier, Grenoble *)
(*                                         2006 *)

(* $Id: NEmptXorLayer.v,v 1.1 2006/07/14 16:52:07 monin Exp $ *)

(* A non-empty Xor layer is defined as a list with n elements, where n # 1 *)
(* (If n = 1, [_] behaves like (X_ns false) ) *)

Require Import List. 
Require Import RelProps.
Require Import Separable.
Require Import LasagnaTerms.
Require Import SortXor.
Require Import XorLayer.

Section sec_n1.

Variable A : Set.

Inductive n1_list : Set :=
  | N1_0 : n1_list
  | N1_2 : A -> A -> (list A) -> n1_list.

Definition list_of_n1 (l: n1_list) : list A :=
  match l with
    | N1_0 => nil
    | N1_2 x y l => x :: y :: l
  end.

Definition n1_of_list (l:list A) : n1_list :=
  match l with
    | nil => N1_0
    | x :: nil => N1_0
    | x :: y :: l => N1_2 x y l
  end.

Lemma list_of_n1_injective : forall l, n1_of_list (list_of_n1 l) = l.
destruct l; reflexivity.
Qed.

End sec_n1.

Implicit Arguments N1_0 [A].
Implicit Arguments N1_2 [A].
Implicit Arguments list_of_n1 [A].
Implicit Arguments n1_of_list [A].
Implicit Arguments list_of_n1_injective [A].

Section sec_struct.

Variable AS : sortable.
Notation A := (|AS|). 

Definition NElnxor : Set := n1_list A.


Definition n1_sorted (l: NElnxor) := sorted (list_of_n1 l).

Theorem n1_sorted_True_False : prop_True_False n1_sorted.
intro l. unfold n1_sorted. apply sorted_True_False. 
Qed.

Definition mk_n1_sorted b l (sl : n1_sorted l) : mayskip_nx ({l:_ |  n1_sorted l}) b
  := MS_no b (exist n1_sorted l sl).

(* *)

Definition SNElnxor := Sinject (Slnxor AS) (list_of_n1_injective (A:=A)).

Definition Sorted_NElnxor :=
  relativize_sortable SNElnxor n1_sorted n1_sorted_True_False.

End sec_struct.

Implicit Arguments n1_sorted [AS].
Implicit Arguments mk_n1_sorted [AS].

Section sec_map.

Variable A B : Set.
Variable f : A -> B.

Definition map_n1_list (l:n1_list A) :=
  match l with
    | N1_0 => N1_0
    | N1_2 x y l => N1_2 (f x) (f y) (List.map f l)
  end.

Theorem map_square :
  forall l, (list_of_n1 (map_n1_list l)) = List.map f (list_of_n1 l).
destruct l as  [ | x y l]; reflexivity.
Qed.

End sec_map.

Implicit Arguments map_n1_list [A B].
Implicit Arguments map_square [A B].

Section sec_struct_map.

Variable A B : sortable.
Variable f : |A| -> |B| .
Hypothesis mf : monot f.

Lemma pres_n1_sorted : forall l, n1_sorted l -> n1_sorted (map_n1_list f l).
destruct l as [ | x y l].
  exact (pres_sorted mf nil).
  exact (pres_sorted mf (x :: y :: l)).
Qed. 

Definition map_SNEl: |Sorted_NElnxor A| -> |Sorted_NElnxor B| :=
   fun s => let (x, Hx) := s in
       exist (fun y => n1_sorted y) (map_n1_list f x) (pres_n1_sorted x Hx).

Theorem map_n1_list_monot : monot map_SNEl.
unfold map_SNEl. (* mandatory -- 2nd order matching pb *)
intro l1; destruct l1 as [l1 Sl1]; intro l2; destruct l2 as [l2 Sl2]; simpl.
intro l1l2; repeat rewrite map_square.
apply (map_Slnxor_monot f mf _ _ l1l2).
Qed.

End sec_struct_map.

Implicit Arguments pres_n1_sorted [A B f].
Implicit Arguments map_SNEl [A B].
Implicit Arguments map_n1_list_monot [A B].


Definition s_map_n1 A B (f: monot_fun A B) :
  monot_fun (Sorted_NElnxor A) (Sorted_NElnxor B) :=
  let (f, mf) := f in MF (map_SNEl f mf) (map_n1_list_monot f mf).
Implicit Arguments s_map_n1 [A B].
