(* AUTEURS / AUTHORS :                          *)
(* Judicaël Courant,   Jean-François Monin      *)
(* VERIMAG, Université Joseph Fourier, Grenoble *)
(*                                         2006 *)

(* $Id: NXLayer.v,v 1.7 2006/07/14 16:50:56 monin Exp $ *)

(* Version avec init_term inclus dans nx_term *)

Require Import Bool. (* Is_true *)
Require Import RelProps.
Require Import Separable.
Require Import Terms.        (* for constants *)
Require Import LasagnaTerms. (* using nx_term *)

(* *)

Definition nat_of_secret (x: secret_const) : nat :=
  match x with
  | KM => 0
  | KEK => 1
  | P => 2
  | KEK2 => 3
  | EXP1 => 4
  end.

Definition secret_of_nat (x: nat) :=
  match x with
  | 0 => KM
  | 1 => KEK
  | 2 => P
  | 3 => KEK2
  | 4 => EXP1
  | _ => EXP1
  end.

Lemma inject_nat_secret: 
  forall (x:secret_const), secret_of_nat (nat_of_secret x) = x.
destruct x; reflexivity. 
Qed.

Definition Ssecret := mksrt_from_nat inject_nat_secret.

(* *)

Definition nat_of_public (x: public_const) : nat :=
  match x with
  | DATA => 0
  | PIN => 1
  | ID => 2
  | IMP => 3
  | K3 => 4
  | ACC => 5
  | KP => 6
  | EXP => 7
  end.

Definition public_of_nat (x: nat) :=
  match x with
  | 0 => DATA
  | 1 => PIN
  | 2 => ID
  | 3 => IMP
  | 4 => K3
  | 5 => ACC
  | 6 => KP
  | 7 => EXP
  | _ => EXP
  end.

Lemma inject_nat_public: 
  forall (x:public_const), public_of_nat (nat_of_public x) = x.
destruct x; reflexivity. 
Qed.

Definition Spublic := mksrt_from_nat inject_nat_public.

(* ------------------------------------------------------------------- *) 

Section sec_nx.

Variable AS : sortable.

Notation A := (|AS|). 
Notation ueqA := (ueq AS). 
Notation ueqA_True_False := (ueq_True_False AS).
Notation ueqA_equal := (ueq_equal AS).
Notation ueqA_refl := (ueq_refl AS).

Notation ultA := (ult AS).
Notation ultA_True_False := (ult_True_False AS).
Notation ultA_ueqA_False := (ult_ueq_False AS).
Notation ultA_trans := (ult_trans AS).
Notation ultA_total := (ult_total AS).

Hint Resolve ueqA_True_False ueqA_equal ueqA_refl : srt_db.
Hint Resolve ultA_True_False ultA_ueqA_False ultA_trans ultA_total : srt_db.

(* ------------------------------------------------------------------ *)
(* Next layer *)

Fixpoint ueq_nx_term b (x: nx_term A b) (y: nx_term A b) {struct x}: Prop :=
  match x, y with
  | NX_sum _ _ x, NX_sum _ _ y => ueqA x y
  | NX_PC _ x, NX_PC _ y => ueq Spublic x y
  | NX_SC _ x, NX_SC _ y => ueq Ssecret x y
  | NX_E _ x1 y1, NX_E _ x2 y2 => ueq_nx_term _ x1 x2 /\ ueq_nx_term _ y1 y2
  | NX_Hash _ x1 y1, NX_Hash _ x2 y2 => ueq_nx_term _ x1 x2 /\ ueq_nx_term _ y1 y2
  | _, _ => False
  end.

Remark uniq_Is_true : forall b, True_False (Is_true b). 
destruct b; auto with sep_db. 
Qed.

Lemma ueq_nx_term_True_False : forall b, rel_True_False (ueq_nx_term b).
red. 
induction x as [bx px|bx sx | | | ]; destruct y as [by py|by sy | | | ]; simpl; intuition.
  exact (ueq_True_False Spublic px py).
  exact (ueq_True_False Ssecret sx sy).
Qed.

Lemma ueq_nx_term_equal : forall b, is_equal (ueq_nx_term b).
red; induction x as [b1 p1 | b1 s1 | b1 it1 a1 | b1 x1 Hx1 y1 Hy1 | b1 x1 Hx1 y1 Hy1]; 
destruct y as [b2 p2 | b2 s2 | b2 it2 a2 | b2 x2 y2| b2 x2 y2]; simpl; 
intro U; reflexivity || false_hyp || idtac.
  (* BIZARRE exact (ueq_public_equal _ _ H) donne
The term "ueq_public_equal p p0 H" has type "p = p0"
 while it is expected to have type "p = p0"
PB d'univers pour = ??
*)
  rewrite (ueq_equal Spublic p1 p2); trivial.
  rewrite (ueq_equal Ssecret s1 s2); trivial.
  rewrite (ueqA_equal a1 a2); trivial.
    rewrite (uniq_True_False (uniq_Is_true b2) it1 it2); trivial.
  rewrite (Hx1 x2); intuition. rewrite (Hy1 y2); intuition.
  rewrite (Hx1 x2); intuition. rewrite (Hy1 y2); intuition.
Qed.

Lemma ueq_nx_term_refl : forall b, reflexive (ueq_nx_term b).
red; induction x; simpl; auto. 
  apply (ueq_refl Spublic). 
  apply (ueq_refl Ssecret). 
  apply ueqA_refl.
Qed.

(* *)

Definition nat_of_nx_term b (u: nx_term A b) : nat :=
  match u with
  | NX_PC _ _ => 0
  | NX_SC _ _ => 1
  | NX_sum _ _ _ => 2
  | NX_E _ _ _ => 3
  | NX_Hash _ _ _ => 4
  end.

Fixpoint ult_nx_term b (u v: nx_term A b) {struct u} : Prop :=
  match u, v with
  | NX_PC _ c1, NX_PC _ c2 => ult Spublic c1 c2
  | NX_SC _ c1, NX_SC _ c2 => ult Ssecret c1 c2
  | NX_sum _ _ x1, NX_sum _ _ x2 => ultA x1 x2
  | NX_E _ u1 v1, NX_E _ u2 v2 =>
      ult_nx_term true u1 u2 \/ ueq_nx_term true u1 u2 /\ ult_nx_term true v1 v2
  | NX_Hash _ u1 v1, NX_Hash _ u2 v2 =>
      ult_nx_term true u1 u2 \/ ueq_nx_term true u1 u2 /\ ult_nx_term true v1 v2
  | u, v  => ult_nat (nat_of_nx_term _ u) (nat_of_nx_term _ v) 
  end.

Theorem ult_ueq_nx_term_False : forall b, incompatibles (ult_nx_term b) (ueq_nx_term b).
generalize (ult_ueq_False Spublic) (ult_ueq_False Ssecret) ultA_ueqA_False; simpl. 
red; induction x; destruct y; simpl; intro; reflexivity || false_hyp || intuition; 
unfold ult_inj in * |- ; simpl in * |- ; eauto. 
Qed.

Theorem ult_nx_term_True_False : forall b, rel_True_False (ult_nx_term b).
generalize ueq_nx_term_True_False ult_ueq_nx_term_False.
intros se sl. red in se. 
red; induction x as [bx px|bx sx | | | ]; destruct y as [by py|by sy | | | ]; simpl; intuition.
  apply (ult_True_False Spublic px py). 
  apply (ult_True_False Ssecret sx sy). 
  apply TF_disj; [auto | apply TF_conj; auto | firstorder].
  apply TF_disj; [auto | apply TF_conj; auto | firstorder]. 
Qed.

Ltac subs_ueq_nx :=
   match goal with
   | [ U: ueq_nx_term ?b ?x ?y |- _ ] =>
        generalize (ueq_nx_term_equal b x y U); let E := fresh in intro E; subst
   end.

(* De derrire les fagots : 125/125 cas auto, soit 100 %, mieux que B ! *)
(* Amliorer efficacit en passant par ult_nat_trans *)
Theorem ult_nx_term_trans : forall b, transitive (ult_nx_term b).
generalize (ult_trans Spublic) (ult_trans Ssecret) ultA_trans.
unfold transitive; simpl.
induction x; destruct y; destruct z; simpl; intuition; repeat subs_ueq_nx; eauto.
Qed.

Hint Resolve (ueq_nx_term_refl true) : ueq_db.

Theorem ult_nx_term_total : forall b, total (ult_nx_term b).
red. 
induction x as [ b1 p1 | b1 s1 | b1 it1 a1 | b1 x1 Hx1 y1 Hy1 | b1 x1 Hx1 y1 Hy1]; 
destruct y as [ b2 p2 | b2 s2 | b2 it2 a2 | b2 x2 y2 | b2 x2 y2]; solve_cmptype.
  case (ult_total Spublic p1 p2); destruct_cmptype.
  case (ult_total Ssecret s1 s2); destruct_cmptype.
  rewrite (uniq_True_False (uniq_Is_true b2) it1 it2); 
     case (ultA_total a1 a2); destruct_cmptype. 
  case (Hx1 x2); destruct_cmptype. 
     case (Hy1 y2); destruct_cmptype. 
  case (Hx1 x2); destruct_cmptype. 
     case (Hy1 y2); destruct_cmptype. 
Qed.

(* *)

Definition Snx_term b :=
   mksrt (ueq_nx_term_True_False b) (ueq_nx_term_equal b) (ueq_nx_term_refl b)
     (ult_nx_term_True_False b) (ult_ueq_nx_term_False b) (ult_nx_term_trans b) (ult_nx_term_total b).

End sec_nx.

Implicit Arguments ueq_nx_term [AS].
Implicit Arguments ult_nx_term [AS].

Implicit Arguments ueq_nx_term_True_False [AS].
Implicit Arguments ueq_nx_term_equal [AS].
Implicit Arguments ueq_nx_term_refl [AS].
Implicit Arguments ult_nx_term_True_False [AS].
Implicit Arguments ult_ueq_nx_term_False [AS].
Implicit Arguments ult_nx_term_trans [AS].
Implicit Arguments ult_nx_term_total [AS].

Section sec_maps.

Variable A B : sortable.
Variable f : |A| -> |B| .
Hypothesis mf : monot f.
Variable b: bool.

(* Warning : this typing of [map_nx f b] as an arg of monot is not inferred *)
Definition map_Snx: |Snx_term A b| -> |Snx_term B b| := map_nx f b. 

(* Pompe dans NXLayer *)
Ltac subs_ueq_nx :=
   match goal with
   | [ U: ueq_nx_term ?b ?x ?y |- _ ] =>
        generalize (ueq_nx_term_equal b x y U); let E := fresh in intro E; subst
   end.

Hint Resolve ueq_nx_term_refl : mon_db. (*marche pas automatiqt*)

(* 23 cases / 25 automatically handled *)
Lemma map_Snx_monot : monot map_Snx.
unfold map_Snx. (* mandatory -- 2nd order matching pb *)
intro x; induction x; intro y; destruct y; simpl in *|-* ; intuition.
   right; intuition. subs_ueq_nx; apply ueq_nx_term_refl. 
   right; intuition. subs_ueq_nx; apply ueq_nx_term_refl. 
Qed.

End sec_maps.

Implicit Arguments map_Snx [A B].
Implicit Arguments map_Snx_monot [A B].

Definition s_map_nx A B (f : monot_fun A B) (b : bool) :
  monot_fun (Snx_term A b) (Snx_term B b) :=
  let (f, mf) := f in MF (map_Snx f b) (map_Snx_monot f mf b).
Implicit Arguments s_map_nx [A B].
