(* AUTEURS / AUTHORS :                          *)
(* Judicaël Courant,   Jean-François Monin      *)
(* VERIMAG, Université Joseph Fourier, Grenoble *)
(*                                         2006 *)

(* $Id: Nat2.v,v 1.5 2006/03/08 13:35:24 monin Exp $ *)

(* 
Alternating definition of natural numbers and basic arithmetic:
  - plus, minus
  - max, conveniently defined as  n + (m - n)

*)
(* ---------------------------------------------------------------------- *)

Inductive alte : Set :=  (* even *)
  | AZ : alte
  | ES : alto -> alte
with alto : Set :=       (* odd *)
  | OS : alte -> alto
.

Scheme e_ind := Induction for alte Sort Prop
with o_ind := Induction for alto Sort Prop.

(* ---------------------------------------------------------------------- *)

(* For internalization *)

Inductive parity : Set :=
  | Pev : parity
  | Pod : parity
.

Definition is_even p : bool :=
  match p with
  | Pev => true
  | Pod => false
  end.

Definition negp p : parity :=
  match p with
  | Pev => Pod
  | Pod => Pev
  end.

Require Import Bool.

Definition is_odd p : bool := negb (is_even p).

Definition alt_of_parity p : Set :=
  match p with
  | Pev => alte
  | Pod => alto
  end.

(* A FAIRE : expérimenter alt_nat au lieu de alte/alto *)
(* 2 versions possibles suivant place de (negp p) *)
Inductive alt_nat : parity -> Set :=
  | A0 : alt_nat Pev
  | AS : forall p, alt_nat (negp p) -> alt_nat p
.


(* ---------------------------------------------------------------------- *)

Fixpoint half_e (e: alte): nat :=
  match e with
  | AZ => 0
  | ES o => S (half_o o)
  end 
with half_o (o : alto): nat :=
  match o with
  | OS e => half_e e
  end.

Fixpoint double_e (n: nat): alte :=
  match n with
  | O => AZ
  | S n => ES (OS (double_e n))
  end.

Definition double_o (n: nat): alto := OS (double_e n).

(* *)

Lemma half_double_e : forall n, half_e (double_e n) = n.
induction n; simpl; auto.
Qed.

Lemma half_double_o : forall n, half_o (double_o n) = n.
induction n; simpl; auto.
Qed.

Definition double_half_e_stmt e := double_e (half_e e) = e.
Definition double_half_o_stmt o := double_o (half_o o) = o.

Definition double_half_AZ : double_half_e_stmt AZ. 
red; reflexivity. 
Defined.

Definition double_half_o_e : forall o : alto, double_half_o_stmt o -> double_half_e_stmt (ES o).
unfold double_half_e_stmt; unfold double_half_o_stmt; simpl; intros.
apply (f_equal ES). assumption. 
Defined.

Definition double_half_e_o : forall e : alte, double_half_e_stmt e -> double_half_o_stmt (OS e).
unfold double_half_e_stmt; unfold double_half_o_stmt; simpl; intros.
unfold double_o. apply (f_equal OS). assumption. 
Defined.

Definition double_half_e : forall e, double_e (half_e e) = e 
:= e_ind double_half_e_stmt double_half_o_stmt double_half_AZ double_half_o_e double_half_e_o.

Definition double_half_o : forall o, double_o (half_o o) = o
:=  o_ind double_half_e_stmt double_half_o_stmt double_half_AZ double_half_o_e double_half_e_o.

(* *)

Fixpoint plus_ee (e1 e2: alte) {struct e1} : alte :=
  match e1 with
  | AZ => e2
  | ES o1 => ES (plus_oe o1 e2)
  end 
with plus_oe (o1 : alto) (e2: alte) {struct o1} : alto :=
  match o1 with
  | OS e1 => OS (plus_ee e1 e2)
  end.

Notation "x +e y" := (plus_ee x y) (at level 50, left associativity).
Notation "x +o y" := (plus_oe x y) (at level 50, left associativity).

(* Semble inutile *)
Fixpoint plus_eo (e1: alte) (o2: alto) {struct e1} : alto :=
  match e1 with
  | AZ => o2
  | ES o1 => OS (plus_oo o1 o2)
  end 
with plus_oo (o1 : alto) (o2: alto) {struct o1} : alte :=
  match o1 with
  | OS e1 => ES (plus_eo e1 o2)
  end.

Lemma double_ee_plus :
  forall n m, plus_ee (double_e n) (double_e m) = double_e (plus n m).
induction n; simpl. 
  reflexivity. 
  intros m. rewrite IHn. reflexivity. 
Qed.

Lemma double_eo_plus :
  forall n m, plus_eo (double_e n) (double_o m) = double_o (plus n m).
induction n; simpl. 
  reflexivity. 
  intros m. rewrite IHn. reflexivity. 
Qed.

Lemma double_oe_plus :
  forall n m, plus_oe (double_o n) (double_e m) = double_o (plus n m).
intros n m. unfold double_o; simpl. rewrite double_ee_plus. reflexivity.
Qed.

Lemma double_oo_plus :
  forall n m, plus_oo (double_o n) (double_o m) = double_e (S (plus n m)).
intros n m. unfold double_o; simpl. apply (f_equal ES). 
  elim n; clear n; simpl. 
    reflexivity. 
    intros n Hn . rewrite Hn. reflexivity. 
Qed.

(* *)

Fixpoint minus_ee (e1 e2: alte) {struct e1} : alte :=
  match e1, e2 with
  | AZ, _ => AZ
  | _, AZ => e1 
  | ES o1, ES o2 => minus_oo o1 o2
  end 
with minus_oo (o1 : alto) (o2: alto) {struct o1} : alte :=
  match o1, o2 with
  | OS e1, OS e2 => minus_ee e1 e2
  end.

Notation "x -e y" := (minus_ee x y) (at level 50, left associativity).
Notation "x -o y" := (minus_oo x y) (at level 50, left associativity).

Lemma double_ee_minus :
  forall n m, minus_ee (double_e n) (double_e m) = double_e (minus n m).
induction n; simpl. 
  reflexivity. 
  destruct m; simpl. 
    reflexivity.
    apply IHn. 
Qed.

Lemma double_oo_minus :
  forall n m, minus_oo (double_o n) (double_o m) = double_e (minus n m).
intros n m; unfold double_o. simpl. apply double_ee_minus.
Qed.


(* *)

(* A FAIRE TODO: automatiser ces preuves faciles *)

Require Arith.

Lemma plus_0_ee : forall e, plus_ee e AZ = e.
intro e; pattern e; rewrite <- double_half_e.
match goal with [|- context f [AZ]] => let g := context f [double_e O] in change g end.
rewrite double_ee_plus. rewrite <- (plus_n_O (half_e e)). reflexivity.
Qed.

Lemma plus_S_ee : forall e1 o2, plus_ee e1 (ES o2) = ES (plus_eo e1 o2).
intros e1 o2.
pattern e1; rewrite <- double_half_e.
pattern o2; rewrite <- double_half_o.
rewrite double_eo_plus. 
repeat match goal with [|- context f [ES (double_o ?x)]] => let g := context f [double_e (S x)] in change g end. 
rewrite double_ee_plus. 
apply (f_equal double_e).
rewrite plus_n_Sm. 
reflexivity.
Qed.

Lemma plus_S_eo : forall e1 e2, plus_eo e1 (OS e2) = OS (plus_ee e1 e2).
intros e1 e2.
pattern e1; rewrite <- double_half_e.
pattern e2; rewrite <- double_half_e.
match goal with [|- context f [OS (double_e ?x)]] => let g := context f [double_o x] in change g end. rewrite double_eo_plus. 
rewrite double_ee_plus. 
reflexivity.
Qed.

(* suivant pas fait -- inutile *)
Definition plus_S_oo_stmt o1 := forall e2, plus_oo o1 (OS e2) = ES (plus_oe o1 e2).

(* inutile *)
Lemma plus_S_oe : forall o1 o2, plus_oe o1 (ES o2) = OS (plus_oo o1 o2).
intros o1 o2.
pattern o1; rewrite <- double_half_o.
pattern o2; rewrite <- double_half_o. simpl. 
rewrite plus_S_ee. 
reflexivity.
Qed.


(* *)

Definition max n m := n + (m - n). 

Lemma sym_max : forall n m, max n m = max m n.
unfold max. induction n; destruct m; simpl; auto with arith.
Qed.

Lemma S_max : forall n m, S(max n m) = max (S n) (S m).
unfold max. simpl. auto with arith.
Qed.

(* *)

Definition max_ee e1 e2 := plus_ee e1 (minus_ee e2 e1).
Definition max_oo o1 o2 := plus_oe o1 (minus_oo o2 o1).


Theorem sym_max_ee :  forall e1 e2,  max_ee e1 e2 = max_ee e2 e1.
intros e1 e2; unfold max_ee.
pattern e1; rewrite <- double_half_e.
pattern e2; rewrite <- double_half_e.
repeat rewrite double_ee_minus. 
repeat rewrite double_ee_plus. 
apply (f_equal double_e). 
exact (sym_max (half_e e1) (half_e e2)).
Qed.


Theorem sym_max_oo :  forall o1 o2,  max_oo o1 o2 = max_oo o2 o1.
intros o1 o2; unfold max_oo.
pattern o1; rewrite <- double_half_o.
pattern o2; rewrite <- double_half_o.
repeat rewrite double_oo_minus. 
repeat rewrite double_oe_plus. 
apply (f_equal double_o). 
exact (sym_max (half_o o1) (half_o o2)).
Qed.



