(* AUTEURS / AUTHORS :                          *)
(* Judicaël Courant,   Jean-François Monin      *)
(* VERIMAG, Université Joseph Fourier, Grenoble *)
(*                                         2006 *)

(* $Id: SortXor.v,v 1.5 2006/07/14 16:50:56 monin Exp $ *)

Require Import List. 
Require Import RelProps.
Require Import Separable. 
Require Import LasagnaTerms. (* using xor_term *)


(* *)

Ltac simpl_f_equal :=
  match goal with
  | [ |- ?f ?x = ?f ?y ] => apply (f_equal f)
  | [ |- ?f ?x1 ?x = ?f ?x1 ?y ] => apply (f_equal (f x1))
  | [ |- ?f ?x1 ?x2 ?x = ?f ?x1 ?x2 ?y ] => apply (f_equal (f x1 x2))
  | [ |- ?f ?x1 ?x2 ?x3 ?x = ?f ?x1 ?x2 ?x3 ?y ] => apply (f_equal (f x1 x2 x3))
  end.

Ltac new_hyp E :=
  let t := type of E in
  match goal with
  | [ _: t |- _ ] => idtac
  | [ |- _ ] => generalize E; intro
  end.

Section sec_sort.

Variable AS : sortable.

Notation A := (|AS|). 
Notation ueqA := (ueq AS). 
Notation ueqA_True_False := (ueq_True_False AS).
Notation ueqA_equal := (ueq_equal AS).
Notation ueqA_refl := (ueq_refl AS).

Notation ultA := (ult AS).
Notation ultA_True_False := (ult_True_False AS).
Notation ultA_ueqA_False := (ult_ueq_False AS).
Notation ultA_trans := (ult_trans AS).
Notation ultA_total := (ult_total AS).

Hint Resolve ueqA_True_False : srt_db.
Hint Resolve ueqA_True_False ueqA_equal ueqA_refl : srt_db.
Hint Resolve ultA_True_False ultA_ueqA_False ultA_trans ultA_total : srt_db.

(* *)

Lemma ns_sep : separable A.
exact (uniq_sep ueqA_True_False ueqA_equal ueqA_refl).
Qed.

Lemma uniq_ueqA : rel_unique ueqA.
red; intros; apply uniq_True_False; auto with srt_db.
Qed.

Lemma ultA_irrefl : irreflexive ultA.
exact (nueq_irrefl ueqA_refl ultA_ueqA_False).
Qed.

Lemma uniq_ultA : rel_unique ultA.
red; intros; apply uniq_True_False; auto with srt_db.
Qed.

(* forward saturation of contextual ultA *)

Definition ultA_black := ultA. 

(* warning: no circuit detection *)
Ltac findmin y c :=
  match goal with
  | [ _:ultA ?x y |- _ ] => findmin x c
  | [ |- _ ] => c y
  end.

Ltac fwd_ultA_add H1 x y :=
  match goal with
  | [ H2:ultA y ?z |- _ ] =>
      fold ultA_black in H2;
      fwd_ultA_add H2 y z;
      new_hyp (ultA_trans _ _ _ H1 H2);
      fwd_ultA_add H1 x y
  | [ |- _ ] => idtac
  end.

Ltac fwd_ultA_sat x :=
  match goal with
  | [ H1:ultA x ?y |- _ ] =>
      fold ultA_black in H1;
      fwd_ultA_add H1 x y;
      fwd_ultA_sat x
  | [ |- _ ] => idtac
  end.

Ltac forward_ultA := 
  match goal with
  | [ _:ultA ?x ?y |- _ ] => findmin x fwd_ultA_sat; forward_ultA
  | [ |- _ ] => unfold ultA_black in * |- 
  end.


(* -------------------------------------------------------------------------- *)

Lemma not_ultA_xy_yx : forall x y, ultA x y -> ultA y x -> False.
intros. apply (ultA_irrefl x). eauto with srt_db.
Qed.

(* amliorer, pas complte *)
Ltac abshyp :=
  match goal with 
  | H:False |- _ => case H
  | H:ultA ?x ?x |- _ => case (ultA_irrefl _ H)
  | H1:ultA ?x ?y, H2: ultA ?y ?x |- _ => case (not_ultA_xy_yx _ _ H1 H2)
 end. 

(* -------------------------------------------------------------------------- *)

Lemma conv_ultA_total :
  forall (T:Type) x y (f g h : A->A->T),
   match ultA_total x y with
   | Clt _ => f x y
   | Ceq _ => g x y
   | Cgt _ => h x y
   end =
   match ultA_total y x with
   | Clt _ => h x y
   | Ceq _ => g x y
   | Cgt _ => f x y
   end.
intros. case (ultA_total x y); case (ultA_total y x);
  intros; reflexivity || subst; abshyp.
Qed.

(* ========================================================================== *)

Definition lnxor := list A.

Fixpoint ueq_lnxor (l1 l2 : lnxor){struct l1} : Prop :=
  match l1, l2 with
  | nil, nil => True
  | x1 :: l1, x2 :: l2 => ueqA x1 x2 /\ ueq_lnxor l1 l2
  | _, _ => False
  end.

Lemma ueq_lnxor_True_False : rel_True_False ueq_lnxor.
red; induction x; destruct y; simpl; intuition.
Qed.

Ltac subst_ueqA :=
  match goal with
  | [U: ueqA ?a ?b |- _] => generalize (ueqA_equal a b U); clear U; intro U; subst
  end.

(* *)
Lemma ueq_lnxor_equal : is_equal ueq_lnxor.
red; induction x; destruct y; simpl; intro; reflexivity || false_hyp || intuition.
subst_ueqA. rewrite (IHx y); auto.
Qed.

Lemma ueq_lnxor_refl : reflexive ueq_lnxor.
red; induction x; simpl; auto with srt_db.
Qed.

Lemma lnxor_sep : separable lnxor.
exact (uniq_sep ueq_lnxor_True_False ueq_lnxor_equal ueq_lnxor_refl).
Defined.

(* Sorted lists *)

Set Implicit Arguments.

Fixpoint sorted1 (x : A) (l : lnxor){struct l} : Prop :=
  match l with
  | nil => True
  | y :: l => ultA x y /\ sorted1 y l 
  end.

Lemma sorted1_True_False : forall x l, True_False (sorted1 x l).
intros x l; generalize x; clear x; elim l; simpl; auto with srt_db sep_db. 
Qed.
Hint Resolve sorted1_True_False : sep_db.

Ltac destruct_sorted :=
  match goal with
  | H : sorted1 ?x nil |- _ => clear H
  | H : sorted1 ?x (?a :: ?l) |- _ => simpl in H; case H; clear H; do 2 intro
  end.

Definition sorted (l : lnxor) : Prop :=
  match l with
  | nil => True
  | x :: l => sorted1 x l 
  end.

Unset Implicit Arguments.

Theorem sorted_True_False : prop_True_False sorted.
red; destruct x; simpl; auto with sep_db. 
Qed.

(* *)

Lemma sorted1_monot :
  forall x y, ultA x y  -> forall l, sorted1 y l -> sorted1 x l.
destruct l; simpl; intuition; eauto with srt_db. 
Qed.

Lemma inv_sorted1 : forall x l, sorted1 x l -> sorted l.
intros x l; case l; clear l; simpl; intuition.
Qed.

Lemma inv_sorted : forall x l, sorted (x :: l) -> sorted l.
exact inv_sorted1.
Qed.

Definition sorted1_black := sorted1.
Ltac forward_sorted :=
  match goal with
  | [ Hs: sorted1 ?y ?l |- _ ] => 
     match goal with
     | [ Hl: ultA ?x y |- _ ] =>
        fold ultA_black in Hl;
        new_hyp (sorted1_monot x y Hl l Hs)
     end;
     forward_sorted
  | [ H: sorted1 ?x ?l |- _ ] => 
     fold sorted1_black in H;
     new_hyp (inv_sorted1 x l H);
     forward_sorted
  | [ |- _ ] => unfold sorted1_black in *; unfold ultA_black in *
  end.

Lemma sorted1_all_ultA : forall l x, sorted1 x l -> forall y, In y l -> ultA x y.
induction l; simpl.
  intros; abshyp.
  intuition; subst; eauto with srt_db.
Qed.

(* unused *)
Remark sorted1_all_diff : forall x l, sorted1 x l -> forall y, In y l -> x<>y.
intros x l s y i; generalize (sorted1_all_ultA l x s y i). 
intros lxy e; subst. apply (ultA_irrefl y lxy).
Qed.

Hint Resolve inv_sorted1 inv_sorted sorted1_monot : sorted_db.

Ltac solve_sorted := try (exact I || auto with sorted_db).

(* -------------------------------------------------------------------------- *)

(* Building sorted list from terms *)
(* First step: insertion of 1 A in a sorted list *)

Fixpoint insert (x : A) (l : lnxor) {struct l} : lnxor :=
  match l with
  | nil => x :: nil
  | y :: l =>
      match ultA_total x y with
      | Clt lxy => x :: y :: l
      | Ceq exy => l
      | Cgt lyx => y :: insert x l
      end
  end.

Remark insert_lt :  forall x y l, ultA x y -> insert x (y :: l) = x :: y :: l.
intros; simpl; case (ultA_total x y); simpl; intros; subst;
  solve [trivial] || abshyp.
Qed.

Remark insert_eq :  forall x y l, x = y -> insert x (y :: l) = l.
intros; simpl; case (ultA_total x y); simpl; intros; subst;
  solve [trivial] || abshyp.
Qed.

Remark insert_gt :  forall x y l, ultA y x -> insert x (y :: l) = y :: insert x l.
intros; simpl; case (ultA_total x y); simpl; intros; subst;
  solve [trivial] || abshyp.
Qed.

Hint Rewrite insert_lt insert_eq insert_gt : insert_db.

Ltac rew_insert := 
  match goal with
  | H : ultA ?x ?y |- _ => rewrite insert_lt; [ idtac | exact H]
  | H : ?x = ?y |- _ => rewrite insert_eq; [ idtac | exact H]
  | H : ultA ?y ?x |- _ => rewrite insert_gt; [ idtac | exact H]
  | |- context f [insert ?a (?a :: ?l)] =>
       generalize (refl_equal a); intro; rewrite insert_eq; [idtac | assumption]
  end.

Ltac clear_equal := 
  match goal with
  | [ e:?x = ?x |- _ ] => clear e; clear_equal
  | [|- _] => idtac
  end.

Ltac case_ultA_total x y := 
  case (ultA_total x y); intro; clear_equal; subst;
    forward_ultA; forward_sorted; repeat rew_insert; clear_equal; trivial. 

Lemma sorted1_insert : forall x l, sorted1 x l -> x :: l = insert x l.
destruct l.
  simpl; intuition.
  intros H. destruct_sorted. rew_insert. trivial. 
Qed. 

Ltac rew_sorted1_insert :=
  match goal with
  | [ H: sorted1 ?a ?l |- context f [ ?a :: ?l ] ] =>
       rewrite (sorted1_insert a l H)
  end.

(* sorted l needed in case of 3, 3, 3 :: 1 :: 5 :: nil *)
Remark insert_invol :  forall x l, sorted l -> insert x (insert x l) = l.
induction l; simpl.
   case (ultA_total x x); intros; reflexivity || abshyp. 
   intro. case_ultA_total x a. 
     rew_sorted1_insert; trivial.
     rewrite IHl; auto with srt_db.  
Qed. 

Ltac case_ultA_tot_simpl x y := case_ultA_total x y; simpl; intuition. 

Remark insert_sorted1 : forall l a x, sorted1 x l -> sorted (insert a (x :: l)).
intro l; elim l; clear l. 
  simpl. intros. case_ultA_tot_simpl a x.
  intros a l Hrec x y s. destruct_sorted. case_ultA_tot_simpl x y.
    case_ultA_tot_simpl x a. generalize (Hrec x a). rew_insert. simpl; auto. 
Qed.

Lemma insert_sorted : forall l x, sorted l -> sorted (insert x l).
destruct l as [|a l]. 
  simpl; trivial.
  intro x. exact (insert_sorted1 l x a).
Qed.

Lemma insert_commut:
  forall x y l, sorted l -> insert x (insert y l) = insert y (insert x l).
induction l; unfold sorted; intro s.
  simpl. case (ultA_total x y); case (ultA_total y x); intros; reflexivity || subst; abshyp.
  forward_sorted. case_ultA_total y a. 
     case_ultA_total x y. case_ultA_total x a. apply sorted1_insert; auto. 
     case_ultA_total x a. symmetry. apply sorted1_insert; auto.
     case_ultA_total x a. rewrite IHl; auto. 
Qed.

(* -------------------------------------------------------------------------- *)

(* 2nd step: iterating insert *)

Set Implicit Arguments.

Fixpoint fold_insert b (t : xor_term A b) {struct t} : lnxor -> lnxor := 
  match t with
  | X_Zero _ => fun l => l
  | X_ns _ _ n => fun l => insert n l
  | X_Xor _ t u => fun l => fold_insert t (fold_insert u l) 
  end.

Definition sort b (t : xor_term A b) : lnxor := fold_insert t nil.

Lemma fold_insert_is_sorted : 
  forall b (t : xor_term A b) l, sorted l -> sorted (fold_insert t l).
induction t; simpl.
  trivial. 
  generalize insert_sorted; eauto. 
  intros; auto. 
Qed.

Lemma sort_is_sorted :
   forall b (t : xor_term A b), sorted (sort t).
intros; unfold sort; apply fold_insert_is_sorted. simpl. trivial. 
Qed.

Definition mk_sorted b (t : xor_term A b) : {l:lnxor | sorted l} :=
  exist _ _ (sort_is_sorted t).

(* *)

Hint Resolve insert_sorted fold_insert_is_sorted sort_is_sorted :
     sorted_db.

Lemma insert_fold_insert_commut_gen :
  forall b (t : xor_term A b) x l, sorted l -> 
  insert x (fold_insert t l) = fold_insert t (insert x l).
induction t; intros; simpl. 
  reflexivity.
  apply insert_commut. assumption. 
  rewrite IHt1. 
    rewrite IHt2. 
      reflexivity. 
      assumption. 
    apply fold_insert_is_sorted; assumption. 
Qed.

Ltac insert_to_front :=
  match goal with
  | [ |- context f [ fold_insert ?t (insert ?x ?l) ] ] =>
       rewrite <- (insert_fold_insert_commut_gen t x l);
         [idtac | auto with sorted_db]
  end.

Lemma sorted_commut_gen :
   forall bt bu (t: xor_term A bt) (u: xor_term A bu) l, sorted l -> 
   fold_insert t (fold_insert u l) = fold_insert u (fold_insert t l).
induction t; simpl. 
  reflexivity.
  intros. apply insert_fold_insert_commut_gen. assumption. 
  intros. rewrite IHt2. 
    apply IHt1. apply fold_insert_is_sorted; assumption.
    assumption. 
Qed.

Lemma fold_insert_invol_gen :
   forall b (t : xor_term A b) l, 
   sorted l -> fold_insert t (fold_insert t l) = l.
induction t; simpl. 
  reflexivity.
  intros. apply insert_invol. assumption. 
  intros. rewrite sorted_commut_gen; 
    repeat (assumption || reflexivity || rewrite IHt1 || rewrite IHt2 || apply fold_insert_is_sorted). 
Qed.

Lemma insert_fold_insert_commut :
   forall b (t : xor_term A b) x, 
   insert x (fold_insert t nil) = fold_insert t (x :: nil).
intros b t x. rewrite insert_fold_insert_commut_gen; simpl; trivial. 
Qed.

Lemma commut :
   forall bt bu (t: xor_term A bt) (u: xor_term A bu), 
   fold_insert t (fold_insert u nil) = fold_insert u (fold_insert t nil).
intros bt bu t u. rewrite sorted_commut_gen; simpl; trivial. 
Qed.

Lemma fold_insert_invol : 
    forall b (t : xor_term A b), fold_insert t (fold_insert t nil) = nil.
intros; apply fold_insert_invol_gen. simpl; trivial. 
Qed.

Unset Implicit Arguments.

Lemma fold_insert_equiv_invol : 
  forall bt bu (t: xor_term A bt) (u: xor_term A bu) l, sorted l->
  fold_insert t (fold_insert u l) = l -> fold_insert t l = fold_insert u l.
intros bt bu t u l s e. pattern l at 1. rewrite <- e.
rewrite fold_insert_invol_gen; simpl; auto.
apply fold_insert_is_sorted. trivial. 
Qed.

Set Implicit Arguments.

Ltac to_invol :=
  match goal with
  | [ |- fold_insert ?t ?l = fold_insert ?u ?l] =>
       apply (fold_insert_equiv_invol _ _ t u l); solve_sorted 
  | [ e : fold_insert ?t ?l = fold_insert ?u ?l |-
       context f [ fold_insert ?t (fold_insert ?u ?x) ]] =>
       case e; solve_sorted; rewrite fold_insert_invol_gen; solve_sorted
  end.


Lemma fold_insert_zero :
  forall bt bu (t: xor_term A bt) (u: xor_term A bu), 
  fold_insert t (fold_insert u nil) = nil -> 
  forall l, sorted l -> fold_insert t (fold_insert u l) = l.
induction l.
  trivial.
  simpl; intros s. forward_sorted. rew_sorted1_insert. repeat insert_to_front. 
     simpl_f_equal. auto. 
Qed.

Lemma fold_insert_monot :
  forall bt bu (t: xor_term A bt) (u: xor_term A bu), 
  fold_insert t nil = fold_insert u nil -> 
  forall l, sorted l -> fold_insert t l = fold_insert u l.
intros. to_invol. apply fold_insert_zero; solve_sorted. to_invol.
Qed.

End sec_sort.

Implicit Arguments lnxor.
Implicit Arguments sort [AS b].
Implicit Arguments mk_sorted [AS b].

Hint Resolve sorted1_True_False : sep_db.
Hint Resolve inv_sorted1 inv_sorted sorted1_monot : sorted_db.
Hint Rewrite insert_lt insert_eq insert_gt : insert_db.
Hint Resolve insert_sorted fold_insert_is_sorted sort_is_sorted : 
     sorted_db.
