(* AUTEURS / AUTHORS :                          *)
(* Judicaël Courant,   Jean-François Monin      *)
(* VERIMAG, Université Joseph Fourier, Grenoble *)
(*                                         2006 *)

(* $Id: XorLayer.v,v 1.5 2006/07/14 16:50:56 monin Exp $ *)

Require Import List. 
Require Import RelProps.
Require Import Separable. 
Require Import LasagnaTerms. (* using xor_term *)

Require Import SortXor.


Section sec_terms.

Variable AS : sortable.

Notation A := (|AS|). 
Notation ueqA := (ueq AS). 
Notation ueqA_True_False := (ueq_True_False AS).
Notation ueqA_equal := (ueq_equal AS).
Notation ueqA_refl := (ueq_refl AS).

Notation ultA := (ult AS).
Notation ultA_True_False := (ult_True_False AS).
Notation ultA_ueqA_False := (ult_ueq_False AS).
Notation ultA_trans := (ult_trans AS).
Notation ultA_total := (ult_total AS).

Hint Resolve ueqA_True_False ueqA_equal ueqA_refl : srt_db.
Hint Resolve ultA_True_False ultA_ueqA_False ultA_trans ultA_total : srt_db.

(* ------------------------------------------------------------------ *)
(* Next layer *)

Notation lnxorA := (lnxor AS).
Definition ueq_lnxor := (ueq_lnxor AS). 

Ltac subst_ueq :=
  match goal with
  | [U: ueqA ?a ?b |- _] => generalize (ueqA_equal a b U); clear U; intro U; subst
  end.

(* ------------------------------------------------------------------ *)

Fixpoint ult_lnxor (l1 l2: lnxorA) {struct l1} : Prop :=
  match l1, l2 with
  | _, nil => False
  | nil, _ :: _ => True
  | x1 :: l1, x2 :: l2 => ultA x1 x2 \/ ueqA x1 x2 /\ ult_lnxor l1 l2
  end.

Lemma ult_lnxor_True_False : rel_True_False ult_lnxor.
generalize ultA_True_False.
red; induction x; destruct y; simpl; eauto with srt_db sep_db. 
  apply TF_disj; auto with srt_db sep_db. 
  intuition. eapply ultA_ueqA_False; eauto. 
Qed.

Lemma ult_ueq_lnxor_False : incompatibles ult_lnxor ueq_lnxor.
generalize ultA_ueqA_False.
red; induction x; destruct y; simpl; intuition; eauto.
Qed.

Lemma ult_lnxor_trans : transitive ult_lnxor.
red; induction x as [|x1 l1 Hl1]; destruct y as [|x2 l2]; destruct z as [|x3 l3]; simpl; intuition; try subst_ueq; intuition.
  assert (ultA x1 x3); [eauto with srt_db | intuition].
  right; eauto.
Qed.

Lemma ult_lnxor_total :  total ult_lnxor.
red. intro l1; elim l1; clear l1.
  intro l2; case l2; simpl; clear l2.
    apply Ceq. reflexivity.
    intros; apply Clt. simpl. trivial.
  intros x1 l1 Hl1 l2; case l2; clear l2.
    apply Cgt. simpl. trivial.
    intros x2 l2. case (ultA_total x1 x2); intro x1x2.
      apply Clt; simpl; intuition. 
      subst. case (Hl1 l2); intro l1l2.
        apply Clt; simpl; intuition. 
        subst. apply Ceq. reflexivity. 
        apply Cgt; simpl. auto with srt_db.
      apply Cgt; simpl. auto.
Qed.

(* *)

Definition Slnxor :=
   mksrt (ueq_lnxor_True_False AS) (ueq_lnxor_equal AS) (ueq_lnxor_refl AS) 
         ult_lnxor_True_False ult_ueq_lnxor_False ult_lnxor_trans ult_lnxor_total.

(* ================================================================== *)
(* Towards using a sig type of sorted lists *)

Definition Sorted_lnxor :=
  relativize_sortable Slnxor (sorted (AS:=AS)) (sorted_True_False AS).

End sec_terms.

Implicit Arguments ueq_lnxor [AS].
Implicit Arguments ult_lnxor [AS].

Implicit Arguments ueq_lnxor_True_False [AS].
Implicit Arguments ueq_lnxor_equal [AS].
Implicit Arguments ueq_lnxor_refl [AS].
Implicit Arguments ult_lnxor_True_False [AS].
Implicit Arguments ult_ueq_lnxor_False [AS].
Implicit Arguments ult_lnxor_trans [AS].
Implicit Arguments ult_lnxor_total [AS].

Section sec_maps.

Variable A B : sortable.
Variable f : |A| -> |B| .
Hypothesis mf : monot f.

Definition map_Slnxor : |Slnxor A| -> |Slnxor B| := List.map f. 

Theorem map_Slnxor_monot : monot map_Slnxor.
unfold map_Slnxor. 
intro l1; induction l1 as [|x1 l1];
intro l2; destruct l2 as [|x2 l2]; simpl in *|-* ; intuition.
right; intuition. subs_ueq; apply ueq_refl.
Qed.


Lemma pres_sorted : forall l, sorted l -> sorted (List.map f l).
destruct l; simpl. 
  intro t; exact t.
  generalize x; clear x; elim l; clear l; simpl. 
    intros x t; exact t. 
    intros y l Hl x (lxy, syl); split.
      apply mf. assumption.
      apply Hl. assumption.
Qed.

Definition map_Sorted_lnxor : |Sorted_lnxor A| -> |Sorted_lnxor B| :=
   fun s => let (x, Hx) := s in
       exist (fun y => sorted y) (List.map f x) (pres_sorted x Hx).

Theorem map_Sorted_lnxor_monot : monot map_Sorted_lnxor.
intro l1; destruct l1 as [l1 Sl1]; intro l2; destruct l2 as [l2 Sl2]. 
exact (map_Slnxor_monot l1 l2). 
Qed.

End sec_maps.

Implicit Arguments map_Slnxor [A B].
Implicit Arguments map_Slnxor_monot [A B].
Implicit Arguments pres_sorted [A B f].
Implicit Arguments map_Sorted_lnxor [A B].
Implicit Arguments map_Sorted_lnxor_monot [A B].

